function [M, M_pos, ROR] = get_M_extended(vehicle, obst_list, d_sensor, t)
% obst_list is a list of all obstacles that intersect with the sensor disk
global database;
PI = 3.1416;
ROR_ANGLE = 0.5; % ROR apply for angles larger than ROR_ANGLE
ROR_active = 1;  % Turn on and off Rule of the Road

M = [];
M_pos = [];
sensor_res = database.sensor_res;

yv = vehicle.state(1,t);  % Vehicle y-coordinate
xv = vehicle.state(2,t);  % Vehicle x-coordinate
psi = vehicle.state(3,t); % Heading

% Create a vector of angles to check
alpha = -PI/2 : sensor_res : PI/2;

% Rules of the road flag
ROR =  0;

% For each obstacle check what angles on the sensor disk that is blocked
for k = 1:length(obst_list)
    
    if obst_list(k).is_rectangle()
        %% Rectangular objects
        first = 0;
        last = PI/2;
       
        v = obst_list(k).get_vertices(t, vehicle.get_d_safe());
        l1 = [v(1,:) v(2,:)];
        l2 = [v(2,:) v(3,:)];
        l3 = [v(3,:) v(4,:)];
        l4 = [v(4,:) v(1,:)];
        lines = [l1; l2; l3; l4];

        % Test variables
        x1 = 0; x2 = 0; y1 = 0; y2 = 0;

        for i = 1:length(alpha)
            theta = mod(alpha(i)+psi, 2*PI);
            x = xv + d_sensor*sin(theta)*cos(alpha(i));  % x-coord on the sensor disk at angle theta
            y = yv + d_sensor*cos(theta)*cos(alpha(i));  % y-coord on the sensor disk at angle theta

            does_intersect = 0;
            for j = 1:4
               out = lineSegmentIntersect(lines(j,:),[yv, xv, y, x]);
               if out.intAdjacencyMatrix == 1
                   does_intersect = 1;
                   break;
               end
            end

            if does_intersect && first == 0
                first = alpha(i);
                y1 = y; x1 = x; % used for plotting
            elseif ~does_intersect && first ~= 0 
                last = alpha(i);
                y2 = y; x2 = x; % used for plotting
                break;
            end
        end
        
        
        
        
    else
        %% Circle objects
        r_obs = obst_list(k).get_radius() + vehicle.get_d_safe();
        c_obs = obst_list(k).state(1:2,t)'; % obstacle center

        first = 0;
        last = PI/2;

        % Test variables
        %x1 = 0; x2 = 0; y1 = 0; y2 = 0;

        for i = 1:length(alpha)
            theta = mod(alpha(i)+psi, 2*PI);
            x = xv + d_sensor*sin(theta)*cos(alpha(i));  % x-coord on the sensor disk at angle theta
            y = yv + d_sensor*cos(theta)*cos(alpha(i));  % y-coord on the sensor disk at angle theta

            dist = dist_to_line([yv, xv], [y, x], c_obs)- 0.01;

            if dist < r_obs && first == 0
                first = alpha(i);
                %y1 = y; x1 = x; % used for plotting
            elseif dist > r_obs && first ~= 0 
                last = alpha(i);
                %y2 = y; x2 = x; % used for plotting
                break;
            end
        end
    end
    

    %% Account for obstacle velocity and update collision angles
    
    v = vehicle.get_speed();
    u = obst_list(k).get_speed();
    psi_obst = obst_list(k).state(3,t);
    
    first_world = mod(psi+first, 2*PI);
    last_world = mod(psi+last, 2*PI);
    
    beta1 = PI - psi_obst + first_world;
    beta2 = PI - psi_obst + last_world;
    
    %beta1 = mod(beta1, PI);
    %beta2 = mod(beta2, PI);
    
    % Set max compensation angle to PI/2
    alpha1 = (u*sin(beta1))/v;
    alpha2 = (u*sin(beta2))/v;
    if abs(alpha1) > 1 
        theta_v1 = sign(alpha1)*PI/2;
    else
        theta_v1 = asin((u*sin(beta1))/v);
    end
    
    if abs(alpha2) > 1 
        theta_v2 = sign(alpha2)*PI/2;
    else
        theta_v2 = asin((u*sin(beta2))/v);
    end
    
    first1 = first + theta_v1;
    last1 = last + theta_v2;
    
    
    %%% NEW method
    if ROR_active == 1
        if last > ROR_ANGLE && first1 < 0
            ROR = 1;
        end

        if last < -ROR_ANGLE && last1 > 0
            last1 = -0.1;
            if last1 < first1
                first1 = -PI/2;
            end
        end
    end
    %%%
    
    
    % Make sure angles still does not exceed +- PI/2
    if first1 > PI/2 || first1 < -PI/2
        first1 = sign(first1)*PI/2;
    end
    
    if last1 > PI/2 || last1 < -PI/2
        last1 = sign(last1)*PI/2;
    end
    
    % Used for ploting 
    %first1_world = mod(psi+first1, 2*PI);
    %last1_world = mod(psi+last1, 2*PI);
    
    %plot([x1 xv], [y1 yv]);
    %plot([x2 xv], [y2 yv]);
    
    % How to plot ray when angle is known, uncompensated
    %plot([xv+10*sin(first_world) xv], [yv+10*cos(first_world) yv]);
    %plot([xv+10*sin(last_world) xv], [yv+10*cos(last_world) yv]);
    
    % Plot compensated rays
    %plot([xv+10*sin(first1_world) xv], [yv+10*cos(first1_world) yv]);
    %plot([xv+10*sin(last1_world) xv], [yv+10*cos(last1_world) yv]);
    
    
    M = [M, [first1 last1]'];
    M_pos = [M_pos, [k first last]'];
    

end
    % Sort M such that the first angles are in rising order
    M = sort(M,2);
 
    % Check that each interval in M is valid
    for j = 1:size(M,2)
       if M(2,j)-M(1,j) < 0.01
           if M(1,j) <= -PI/2 + 0.01
               M(2,j) = M(2,j) + 0.01;
           else
               M(1,j) = M(1,j) - 0.01;
           end
       end
    end
    % Merge overlapping areas
    i = 1;
    len = size(M,2);
    while i < len
        if (M(2,i) >= M(1,i+1)) || (abs(M(2,i) - M(1,i+1))< 0.1)
            %if len > i + 1
            %    M = [[M(1,i) M(2,i+1)]' M(:,i+2:len)];
            %else
                if  M(2,i) >= M(2,i+1)
                    M(2,i+1) = M(2,i);
                end
                if i > 1
                    M = [M(:,1:i-1) [M(1,i) M(2,i+1)]' M(:,i+2:len)];
                else
                    M = [[M(1,i) M(2,i+1)]' M(:,i+2:len)]; 
                end              
            %end
            i = 0;
        end
        i = i + 1;
        len = size(M,2);
    end
    
    % Make sure first value of M is -PI/2
    if ~(M(1,1) < -PI/2 + 0.01)
        M = [-[PI/2 PI/2-0.001]' M];
    end
    % Make sure last value of M is PI/2
    if M(end,end) ~= PI/2
        M = [M [PI/2-0.001 PI/2]'];
    end
    
end