classdef kinematic_agent < handle
    % General class for handling obstacles.
    properties
        state = zeros(4,1);    % x y psi r
        state_dot = zeros(4,1);
        psi_error=0;
        psi_d = 0;
        V_d = [0;0];
        finished = 0;
    end
    
    properties (Access = protected)
        time_step = 1;
        sample_count = 1;
        
        target = [0 0]
        radius = 1;
        d_safe = 0;
        max_surge_speed = 1;
        min_surge_speed = 0;
        u = 1;
        max_yaw_rate = 1;      
        Kp = 15;
        r_prev = 0;
        error_prev = 0;
        max_acc = 0.05;
    end
    
    methods (Access = public)
        function obj = kinematic_agent(time, radius, d_safe, target, max_surge_speed, max_yaw_rate)
            if(nargin>0)
                % Initialize parameter vectors
                obj.sample_count = time.get_sample_count();
                obj.time_step = time.get_time_step;
                
                obj.state = zeros(4,obj.sample_count);
                obj.state_dot = zeros(4,obj.sample_count);
                
                obj.psi_error=zeros(1,obj.sample_count);
                obj.psi_d = zeros(1,obj.sample_count);
                
                obj.radius = radius; 
                obj.d_safe = d_safe;
                
                obj.target = target;
            end
            
            if nargin>4
                obj.max_surge_speed = max_surge_speed;
                obj.u = max_surge_speed;
                obj.min_surge_speed = 0.4*max_surge_speed;
                obj.max_yaw_rate = max_yaw_rate;
            end
        end
        
        %%%
        % Get and Set functions
        %%%
        
        function radius = get_radius(obj)
            radius = obj.radius;
        end
        
        function d_safe = get_d_safe(obj)
            d_safe = obj.d_safe;
        end
        
        function set_max_surge_speed(obj, new_speed)
            obj.max_surge_speed = new_speed;
            obj.u = new_speed;
        end
        
        function max_speed = get_max_speed(obj)
            max_speed = obj.max_surge_speed;
        end
        
        function speed = get_speed(obj)
            speed = obj.u;
        end
        
        function min_speed = get_min_speed(obj)
            min_speed = obj.min_surge_speed;
        end
        
        function set_max_yaw_rate(obj, new_max_yaw_rate)
            obj.max_yaw_rate = new_max_yaw_rate;
        end
        
        function max_yaw_rate = get_max_yaw_rate(obj)
            max_yaw_rate = obj.max_yaw_rate;
        end
        
        function sample_count = get_sample_count(obj)
            sample_count = obj.sample_count;
        end
        
        function psi_error = get_psi_error(obj)
            psi_error = obj.psi_error;
        end
        
        function target = get_target(obj)
            target = obj.target;
        end
        
        function set_target(obj, new_target)
            obj.target = new_target;
        end
        
        function rectangle = is_rectangle(obj)
            rectangle = 0;
        end
        
        function max_acc = get_max_acc(obj)
            max_acc = obj.max_acc;
        end
        
        
        %%%
        % Initial conditions
        %%%
        
        function set_initial_conditions(obj, initial_state, initial_state_dot)
            obj.state = diag(initial_state)*ones(4,obj.sample_count);
            obj.state_dot = diag(initial_state_dot)*ones(4, obj.sample_count);
        end
        
        %%%
        % Next step 
        %%%
        function calculate_next_state(obj, current_time, psi_d, u)
            
            obj.psi_d(current_time)=psi_d;
                
            obj.state(4,current_time) = obj.heading_controller(current_time,psi_d);

            obj.update_state_dot(current_time, u);
            
            obj.state(:,current_time+1) = euler2(obj.state_dot(:,current_time), obj.state(:,current_time), obj.time_step);
            obj.state(3,current_time+1) = mod(obj.state(3,current_time+1),2*pi); %Ensure that heading angle is betweeen 0 and 2pi
        end        
                
        function instant = get_instant_states(obj, time_instant)
            instant.state = obj.state(:, time_instant);
            instant.state_dot = obj.state_dot(:, time_instant-1);
        end
        
        %%%
        % Simple heading controller
        %%%
        function r = heading_controller(obj, current_time, psi_d)          
            psi = obj.state(3,current_time);
           
            error = angle_to_pi_pi(psi-psi_d);
            obj.psi_error(current_time) = error;          
            
            r=-obj.Kp*(error-obj.error_prev) + obj.r_prev;
            obj.r_prev=r;
            
            r=min(r, obj.max_yaw_rate);
            r=max(r, -obj.max_yaw_rate);
            
            obj.error_prev=error;
        end
        
        %%%
        % Return center of senordisk
        %%%
        function c = get_sensor_disk_center(obj, d_sensor, t)
            r = d_sensor/2;
            y = obj.state(2,t) + r*sin(obj.state(3,t));
            x = obj.state(1,t) + r*cos(obj.state(3,t));
            c = [x,y];
        end
        
        %%%
        % Returns the distance to a point
        %%%
        function distance = get_distance_to_point(obj, current_time, point)
            current_position = obj.state(1:2,current_time);
            d = sqrt((point(1)-current_position(1))^2+(point(2)-current_position(2))^2);
            distance = d-obj.get_radius;
        end
        
        
        %%%
        % Returns the safe region circle in the world coord
        %%%
        function safe_region = get_obj_region(obj, current_position, d_safe)
            ang=0:0.01:2*pi;
            x_safe=(obj.radius+d_safe)*cos(ang);
            y_safe=(obj.radius+d_safe)*sin(ang);
            safe_region = [current_position(2) + y_safe; current_position(1) + x_safe];
        end
        
        %%%
        % Plots the obstacle into the figure with handle figure_handle
        %%%
        function plot_agent(obj,current_time, figure_handle, lineWidthSmall, lineWidthBig, d_safe, color)
            figure(figure_handle);
            
            current_position = obj.state(1:2,current_time);
            current_velo = obj.state_dot(1:2,current_time-1);
            Vpref = obj.V_d;
            
            % Plot obstacle center
           % hObstacle = plot(current_position(2), current_position(1),'mx');
           % set(hObstacle,'LineWidth', lineWidthBig);
           
            % Plot obstacle limit
            ang=0:0.01:2*pi;
            xp=obj.radius*cos(ang);
            yp=obj.radius*sin(ang);
            
            hAgentLim = plot(current_position(2)+yp,current_position(1)+xp,'Color', color(1)); hold on;
            set(hAgentLim,'LineWidth', lineWidthBig);
            
            if ~obj.finished()
                q = quiver(current_position(2), current_position(1), Vpref(2), Vpref(1)); hold on
                q.Color = 'r'; q.MaxHeadSize = 1; q.LineWidth = 0.75;
            end
            
            q = quiver(current_position(2), current_position(1), current_velo(2), current_velo(1)); hold on;
            q.Color = 'black'; q.MaxHeadSize = 1; q.LineWidth = 0.75;

            % Plot safe distance
            if color(2) ~= 'w'
                safe_region = obj.get_obj_region(current_position, d_safe);
                hSafeDistance = plot(safe_region(1,:), safe_region(2,:),'Color', color(2), 'Linestyle', '--'); hold on;
                set(hSafeDistance,'LineWidth', lineWidthSmall);
            end

        end

    end
    
    methods (Access=protected)
        function update_state_dot(obj, current_time, u)
            psi = obj.state(3,current_time);
            r = obj.state(4,current_time);
            deltaU = round(abs(obj.u - u),4);
            
            if deltaU == 0
                obj.u = min(u, obj.get_max_speed());
            elseif deltaU >= obj.max_acc
                obj.u = min(obj.u - sign(obj.u - u)*obj.max_acc, obj.get_max_speed());
            end
            if u == 0 && ~round(obj.u - obj.max_acc,3)
                obj.u = 0;
            end
            
            xDot = obj.u*cos(psi);
            yDot = obj.u*sin(psi);
            psiDot = r;
            %uDot = 0;
            rDot = 0;

            obj.state_dot(:,current_time) = [xDot; yDot; psiDot; rDot];
        end        
    end 
    
end

